/* UR2.PRG
This Gauss procedure computes the empirical application reported in 
"Least Squares Forecast Averaging" by Bruce E. Hansen
March 2006
*/

load dat[566,3]=dat.txt;
y=dat[.,1]; 	@ unemployment rate @
x=dat[.,2] - dat[.,3];  @ 5 Year Rate minus Fed Funds Rate @
n=rows(y);

@ Plot Data @
library pgraph;
graphset;
pqgwin many;
_pdate="";
let _pmcolor = 0 0 0 0 0 0 0 0 15;
fonts("complex simgrma");
_pcolor={9 5 13 1};
let _pltype= 6 6;
xtics(1960,2005,5,5);
xt=seqa(1959,1/12,n);
_plegstr="Unemployment Rate\000Interest Rate (5YR-FF)";
_plegctl=1;
xy(xt,y~x);


k=6;
m=4;
t0=120;
let hs = 1 3 6 12;

format 12,4;
output file = ur2.out reset;
"MSE by Horizon, relative to BIC Selection";
"      Horizon      Equal       WBIC       MMA";
for hi (1,rows(hs),1);
h=hs[hi];
t=n-t0-h;
fe=zeros(t,4);
for i (1,t,1);
  yi=y[1:t0+i];
  xi=x[1:t0+i,.];
  {f,aicw,bicw}=w_forecast(yi,xi,h,k,m);
  vf=vec(f);
  mmaw=mma(yi,xi,h,k,m);
  mmaf=vec(mmaw)'vf;
  ef=meanc(vf);
  aicf=vf[maxindc(vec(bicw))];
  bicf=vec(bicw)'vf;
  yf=y[t0+i+h];
  fe[i,1]=aicf-yf;
  fe[i,2]=ef-yf;
  fe[i,3]=bicf-yf;
  fe[i,4]=mmaf-yf;
endfor;
mse=meanc(fe.^2);
mse=mse[2 3 4] /mse[1];
h~mse';
endfor;

@ Calculated MMA and BIC weights for full sample estimates @
h=6;
mmaw=mma(y,x,h,k,m);
{f,aicw,bicw}=w_forecast(y,x,h,k,m);
bicw=bicw.*(bicw .> .0001);
mmaw=mmaw.*(mmaw .> .0001);
"";"";
"Full Sample BIC Weights, h= " h;
bicw;
"";
"Full Sample MMA Weights, h= " h;
mmaw;

output off;


/**************************************************
PROC AR_FORECAST

This procedure calculates a least-squares h-step-ahead forecast of a target variable y
given k lags of y (k>0) and m lags of a scalar variable z (m>=0).

The forecast, AIC, and BIC are computed.

The format is
{yf,aic,bic} = ar_forecast(y,z,y,h,k,m);
where
yf  = point forecast
aic = aic of fitted model
bic = bic of fitted model
y   = target variable
z   = predictor variable
h   = forecast horizon (h>=1)
k   = ar lags (k>=1)
m   = lags of z (m>=0)
****************************************************/

proc (3) = ar_forecast(y,z,h,k,m);
local n,x,j,yy,beta,xf,yf,e,bic,aic,km;
  n=rows(y);
  km=maxc(k|m);
  x=ones(n-km-h+1,1);
  for j (1,k,1);
    x=x~y[km-j+1:n-j-h+1];
  endfor;
  for j (1,m,1);
    x=x~z[km-j+1:n-j-h+1];
  endfor;
  yy=y[km+h:n];
  beta=yy/x;
  xf=1|rev(y[n-k+1:n]);
  if m>0; xf=xf|rev(z[n-m+1:n]); endif;
  yf=xf'beta;
  e=yy-x*beta;
  n=rows(e);
  bic=n*ln((e'e)/n)+ln(n)*(k+m+1);
  aic=n*ln((e'e)/n)+2*(k+m+1);
retp(yf,aic,bic);
endp;

/**************************************************
PROC W_FORECAST

This procedure calculates a vector of least-squares h-step-ahead forecasts 
of a target variable y given j lags of y and l lags of a scalar variable z.

All models are computed for j between 1 and k, and l between 0 and m.  
Thus (m+1)*k models are computed

The forecasts, AIC weights, and BIC weights are computed.

The format is
{f,aic,bic} = w_forecast(y,z,h,k,m);
where
f   = vector of point forecasts
aic = vector of aic weights
bic = vector of bic weights
y   = target variable
z   = predictor variable
h   = forecast horizon (h>=1)
k   = maximum ar lags (k>=1)
m   = maximum lags of z (m>=0)
****************************************************/

proc (3) = w_forecast(y,z,h,k,m);
local f,aic,bic,j,fj,aj,bj;
  f=zeros(k,m+1);
  aic=zeros(k,m+1);
  bic=zeros(k,m+1);
  for j1 (1,k,1);
  for j2 (0,m,1);
    {fj,aj,bj}=ar_forecast(y,z,h,j1,j2);
    f[j1,1+j2]=fj;
    aic[j1,1+j2]=aj;
    bic[j1,1+j2]=bj;
  endfor;
  endfor;
  aic=aic-minc(vec(aic));
  aic=exp(-aic/2);
  aic=aic./sumc(vec(aic));
  bic=bic-minc(vec(bic));
  bic=exp(-bic/2);
  bic=bic./sumc(vec(bic));
retp(f,aic,bic);
endp;

/**************************************************
PROC MMA

This procedure calculates the Mallows Model Average (MMA) weights
for the forecasting regression of h-step-ahead y on j lags of y and 
l lags of a scalar variable z, where j varies between 1 and k, 
and l between 0 and m. 

The forecasts, AIC weights, and BIC weights are computed.

The format is
w   = w_forecast(y,z,h,k,m);
where
w   = vector of mma weights
y   = target variable
z   = predictor variable
h   = forecast horizon (h>=1)
k   = maximum ar lags (k>=1)
m   = maximum lags of z (m>=0)
****************************************************/

proc mma(y,z,h,k,m);
local km,n,nm,x,j,j1,j2,s,bj,yy,xx,sxy,bbeta,ee,sig,a1,a2,alpha0,w,u1,u2,u3,u4,ret;
  n=rows(y);
  km=maxc(k|m);
  x=ones(n-km-h+1,1);
  for j (1,k,1);
    x=x~y[km-j+1:n-j-h+1];
  endfor;
  for j (1,m,1);
    x=x~z[km-j+1:n-j-h+1];
  endfor;
  yy=y[km+h:n];
  xx=x'x;
  sxy=x'yy;
  nm=k*(m+1);
  bbeta=zeros(k+1+m,nm);
  a2=zeros(nm,1);
  for j1 (1,k,1);
  for j2 (0,m,1);
   s=seqa(1,1,j1+1);
   if j2>0; s=s|seqa(k+2,1,j2); endif;
   bj=sxy[s]/xx[s,s];
   bbeta[s,(j1-1)*(m+1)+j2+1]=bj;
   a2[(j1-1)*(m+1)+j2+1]=1+j1+j2;
  endfor;
  endfor;
  ee=yy-x*bbeta;
  sig=sumc(ee[.,nm].^2)/(n-k-1-m);
  a1=(ee'ee);
  a2=-a2*sig;
  alpha0=ones(nm,1)/nm;
  {w,u1,u2,u3,u4,ret}=QProg(alpha0,a1,a2,ones(1,nm),1,0,0,0~1);
  w=w.*(w.>0);
  w=w/sumc(w);
  w=reshape(w,k,m+1);
retp(w);
endp;
